"""
* Copyright (c) 2015-2017 The Hyve B.V.
* This code is licensed under the GNU General Public License,
* version 3.
"""
import requests
import getpass

CLIENT_ID = 'glowingbear-js'


class TransmartAPIBase:
    """ Connect to tranSMART using Python. """

    def __init__(self, host, user=None, password=None, print_urls=False):

        self.host = host
        self.user = user or input("Username: ")
        self.print_urls = print_urls

        self.access_token = self._get_access_token(password)

    def _get_access_token(self, password):
        r = requests.post("{}/oauth/token".format(self.host),
                          params={'grant_type': 'password',
                                  'client_id': CLIENT_ID,
                                  'username': self.user,
                                  'password': password or getpass.getpass("Password: ")
                                  })
        r.raise_for_status()

        return r.json().get('access_token')

    def get_observations(self, study=None, patientSet=None, as_dataframe=True):
        raise NotImplementedError

    def get_patients(self, study=None, patientSet=None, as_dataframe=True):
        raise NotImplementedError

    def get_studies(self, as_dataframe=True):
        raise NotImplementedError

    def get_concepts(self, study):
        raise NotImplementedError

    def get_hd_node_data(self, study, node_name, projection='all_data', genes=None):
        raise NotImplementedError



"""
* Copyright (c) 2015-2017 The Hyve B.V.
* This code is licensed under the GNU General Public License,
* version 3.
"""

import json


class Query:
    """ Utility to build queries for transmart v2 api. """

    def __init__(self, handle=None, method='GET', params=None, hal=False,
                 in_study=None, in_patientset=None, in_concept=None, in_gene_list=None,
                 in_transcript_list=None, operator="and"):
        self.handle = handle
        self.method = method
        self.hal = hal
        self._params = params or {}
        self.json = None

        # Operator for constraints, default is and
        self.operator = operator

        # Subject constraints
        self.in_study = StudyConstraint(in_study)
        self.in_patientset = PatientSetConstraint(in_patientset)
        self.set_concept_constraint(in_concept)

        # Biomarker constraints
        self.in_gene_list = GenesConstraint(in_gene_list)
        self.in_transcript_list = TranscriptConstraint(in_transcript_list)

    @property
    def params(self):
        return self._params

    @params.getter
    def params(self):
        self._params.update(self.get_constraints())
        self._params.update(self.get_biomarker_constraints())
        return self._params

    @property
    def headers(self):
        return {'Accept': 'application/{};charset=UTF-8'.format('hal+json' if self.hal else 'json')}

    def get_constraints(self):
        constraints = []
        for c in (self.in_study, self.in_patientset, self.in_concept):
            if isinstance(c, list):
                for v in c:
                    constraints.append(v.get_constraint())
            elif c.value:
                constraints.append(c.get_constraint())

        if len(constraints) > 1:
            constraints = {'constraint': json.dumps({"type" : self.operator, "args": constraints})}
        elif len(constraints) == 1:
             if self.handle == "/v2/patient_sets":
                 constraints = {'constraint': constraints[0]}
             else:
                 constraints = {'constraint': json.dumps(constraints[0])}
        else:
            constraints = {}
        return constraints

    def get_biomarker_constraints(self):
        constraints = ''.join([str(c) for c in (self.in_transcript_list, self.in_gene_list) if c.value])

        if constraints:
            return {'biomarker_constraint': constraints}
        else:
            return {}

    def __repr__(self):
        return "<Query(handle={}, method={}, params={})>".format(self.handle,
                                                                 self.method,
                                                                 self.params)

    def set_concept_constraint(self, in_concept):
        if isinstance(in_concept, list):
            self.in_concept = []
            for concept_constraint in in_concept:
                if "\\" in concept_constraint:
                    self.in_concept.append(ConceptPathConstraint(concept_constraint))
                else:
                    self.in_concept.append(ConceptCodeConstraint(concept_constraint))
        else:
            if in_concept and "\\" in in_concept:
                self.in_concept = ConceptPathConstraint(in_concept)
            else:
                self.in_concept = ConceptCodeConstraint(in_concept)



class Constraint:
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return json.dumps({"type": self.type_, self.val_name: self.value})

    def get_constraint(self):
        return {"type": self.type_, self.val_name: self.value}


class StudyConstraint(Constraint):
    type_ = 'study_name'
    val_name = 'studyId'


class PatientSetConstraint(Constraint):
    type_ = 'patient_set'
    val_name = 'patientSetId'


class ConceptPathConstraint(Constraint):
    type_ = 'concept'
    val_name = 'path'

class ConceptCodeConstraint(Constraint):
    type_ = 'concept'
    val_name = 'conceptCode'


class BiomarkerConstraint(Constraint):

    def __str__(self):
        return json.dumps({"type": "biomarker",
                           "biomarkerType": self.type_,
                           "params": {"names": self.value}})


class TranscriptConstraint(BiomarkerConstraint):
    type_ = 'transcripts'


class GenesConstraint(BiomarkerConstraint):
    type_ = 'genes'



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: highdim.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='highdim.proto',
  package='highdim',
  syntax='proto2',
  serialized_pb=_b('\n\rhighdim.proto\x12\x07highdim\"W\n\rHighDimHeader\x12\x1d\n\x05\x61ssay\x18\x01 \x03(\x0b\x32\x0e.highdim.Assay\x12\'\n\ncolumnSpec\x18\x02 \x03(\x0b\x32\x13.highdim.ColumnSpec\"n\n\nColumnSpec\x12\x0c\n\x04name\x18\x01 \x02(\t\x12,\n\x04type\x18\x02 \x02(\x0e\x32\x1e.highdim.ColumnSpec.ColumnType\"$\n\nColumnType\x12\n\n\x06\x44OUBLE\x10\x01\x12\n\n\x06STRING\x10\x02\"\x98\x01\n\x05\x41ssay\x12\x0f\n\x07\x61ssayId\x18\x01 \x02(\x03\x12\x11\n\tpatientId\x18\x02 \x02(\t\x12\x16\n\x0esampleTypeName\x18\x03 \x01(\t\x12\x15\n\rtimepointName\x18\x04 \x01(\t\x12\x16\n\x0etissueTypeName\x18\x05 \x01(\t\x12\x10\n\x08platform\x18\x06 \x01(\t\x12\x12\n\nsampleCode\x18\x07 \x01(\t\"L\n\x03Row\x12\r\n\x05label\x18\x01 \x02(\t\x12\x11\n\tbioMarker\x18\x02 \x01(\t\x12#\n\x05value\x18\x03 \x03(\x0b\x32\x14.highdim.ColumnValue\";\n\x0b\x43olumnValue\x12\x17\n\x0b\x64oubleValue\x18\x01 \x03(\x01\x42\x02\x10\x01\x12\x13\n\x0bstringValue\x18\x02 \x03(\tB3\n\"org.transmartproject.rest.protobufB\rHighDimProtos')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)



_COLUMNSPEC_COLUMNTYPE = _descriptor.EnumDescriptor(
  name='ColumnType',
  full_name='highdim.ColumnSpec.ColumnType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='DOUBLE', index=0, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='STRING', index=1, number=2,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=189,
  serialized_end=225,
)
_sym_db.RegisterEnumDescriptor(_COLUMNSPEC_COLUMNTYPE)


_HIGHDIMHEADER = _descriptor.Descriptor(
  name='HighDimHeader',
  full_name='highdim.HighDimHeader',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='assay', full_name='highdim.HighDimHeader.assay', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='columnSpec', full_name='highdim.HighDimHeader.columnSpec', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=26,
  serialized_end=113,
)


_COLUMNSPEC = _descriptor.Descriptor(
  name='ColumnSpec',
  full_name='highdim.ColumnSpec',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='highdim.ColumnSpec.name', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='type', full_name='highdim.ColumnSpec.type', index=1,
      number=2, type=14, cpp_type=8, label=2,
      has_default_value=False, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _COLUMNSPEC_COLUMNTYPE,
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=115,
  serialized_end=225,
)


_ASSAY = _descriptor.Descriptor(
  name='Assay',
  full_name='highdim.Assay',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='assayId', full_name='highdim.Assay.assayId', index=0,
      number=1, type=3, cpp_type=2, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='patientId', full_name='highdim.Assay.patientId', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='sampleTypeName', full_name='highdim.Assay.sampleTypeName', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='timepointName', full_name='highdim.Assay.timepointName', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='tissueTypeName', full_name='highdim.Assay.tissueTypeName', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='platform', full_name='highdim.Assay.platform', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='sampleCode', full_name='highdim.Assay.sampleCode', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=228,
  serialized_end=380,
)


_ROW = _descriptor.Descriptor(
  name='Row',
  full_name='highdim.Row',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='label', full_name='highdim.Row.label', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='bioMarker', full_name='highdim.Row.bioMarker', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='value', full_name='highdim.Row.value', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=382,
  serialized_end=458,
)


_COLUMNVALUE = _descriptor.Descriptor(
  name='ColumnValue',
  full_name='highdim.ColumnValue',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='doubleValue', full_name='highdim.ColumnValue.doubleValue', index=0,
      number=1, type=1, cpp_type=5, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=_descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\020\001'))),
    _descriptor.FieldDescriptor(
      name='stringValue', full_name='highdim.ColumnValue.stringValue', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=460,
  serialized_end=519,
)

_HIGHDIMHEADER.fields_by_name['assay'].message_type = _ASSAY
_HIGHDIMHEADER.fields_by_name['columnSpec'].message_type = _COLUMNSPEC
_COLUMNSPEC.fields_by_name['type'].enum_type = _COLUMNSPEC_COLUMNTYPE
_COLUMNSPEC_COLUMNTYPE.containing_type = _COLUMNSPEC
_ROW.fields_by_name['value'].message_type = _COLUMNVALUE
DESCRIPTOR.message_types_by_name['HighDimHeader'] = _HIGHDIMHEADER
DESCRIPTOR.message_types_by_name['ColumnSpec'] = _COLUMNSPEC
DESCRIPTOR.message_types_by_name['Assay'] = _ASSAY
DESCRIPTOR.message_types_by_name['Row'] = _ROW
DESCRIPTOR.message_types_by_name['ColumnValue'] = _COLUMNVALUE

HighDimHeader = _reflection.GeneratedProtocolMessageType('HighDimHeader', (_message.Message,), dict(
  DESCRIPTOR = _HIGHDIMHEADER,
  __module__ = 'highdim_pb2'
  # @@protoc_insertion_point(class_scope:highdim.HighDimHeader)
  ))
_sym_db.RegisterMessage(HighDimHeader)

ColumnSpec = _reflection.GeneratedProtocolMessageType('ColumnSpec', (_message.Message,), dict(
  DESCRIPTOR = _COLUMNSPEC,
  __module__ = 'highdim_pb2'
  # @@protoc_insertion_point(class_scope:highdim.ColumnSpec)
  ))
_sym_db.RegisterMessage(ColumnSpec)

Assay = _reflection.GeneratedProtocolMessageType('Assay', (_message.Message,), dict(
  DESCRIPTOR = _ASSAY,
  __module__ = 'highdim_pb2'
  # @@protoc_insertion_point(class_scope:highdim.Assay)
  ))
_sym_db.RegisterMessage(Assay)

Row = _reflection.GeneratedProtocolMessageType('Row', (_message.Message,), dict(
  DESCRIPTOR = _ROW,
  __module__ = 'highdim_pb2'
  # @@protoc_insertion_point(class_scope:highdim.Row)
  ))
_sym_db.RegisterMessage(Row)

ColumnValue = _reflection.GeneratedProtocolMessageType('ColumnValue', (_message.Message,), dict(
  DESCRIPTOR = _COLUMNVALUE,
  __module__ = 'highdim_pb2'
  # @@protoc_insertion_point(class_scope:highdim.ColumnValue)
  ))
_sym_db.RegisterMessage(ColumnValue)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('\n\"org.transmartproject.rest.protobufB\rHighDimProtos'))
_COLUMNVALUE.fields_by_name['doubleValue'].has_options = True
_COLUMNVALUE.fields_by_name['doubleValue']._options = _descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\020\001'))
# @@protoc_insertion_point(module_scope)



import setuptools
import os
import re


VERSIONFILE=os.path.join('transmart', '__init__.py')
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    version_string = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in {}.".format(VERSIONFILE,))

with open("requirements.txt", 'r') as f:
    required_packages = f.read().splitlines()

if os.environ.get('READTHEDOCS') == 'True':
    for dependency in ['pandas']:
        for package in required_packages:
            if package.startswith(dependency):
                required_packages.remove(package)


setuptools.setup(
    name="transmart",
    version=version_string,
    url="https://www.github.com/thehyve/transmart-api-client-py/",

    maintainer="developers@thehyve.nl",
    maintainer_email="developers@thehyve.nl",

    description="An python client for communicating with the transmart rest api.",

    packages=setuptools.find_packages(exclude=['tests', 'tests.*']),
    include_package_data=True,

    keywords=['transmart', 'rest', 'api', 'data', 'science'],

    download_url='https://github.com/thehyve/transmart-api-client-py/tarball/{}/'.format(version_string),

    install_requires=required_packages,

    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
)



from pandas.io.json import json_normalize


def _format_observations(observations_result):
    output_cells = []
    indexed_dimensions = []
    inline_dimensions = []

    for dimension in observations_result['dimensionDeclarations']:
        if 'inline' in dimension:
            inline_dimensions.append(dimension)
        else:
            indexed_dimensions.append(dimension)

    for dimension in indexed_dimensions:
        dimension['values'] = observations_result['dimensionElements'][dimension['name']]

    for cell in observations_result['cells']:
        output_cell = {}

        i = 0
        for index in cell['dimensionIndexes']:
            if index is not None:
                output_cell[indexed_dimensions[i]['name']] = \
                    indexed_dimensions[i]['values'][int(index)]
            i += 1

        i = 0
        for index in cell['inlineDimensions']:
            output_cell[inline_dimensions[i]['name']] = index
            i += 1

        if 'stringValue' in cell:
            output_cell['stringValue'] = cell['stringValue']
        if 'numericValue' in cell:
            output_cell['numericValue'] = cell['numericValue']
        output_cells.append(output_cell)
    return output_cells


class ObservationSet:
    """ Class to represent observation sets from tranSMART API v2 """

    def __init__(self, json):
        self.json = json
        try:
            self.dataframe = json_normalize(_format_observations(self.json))
        except:
            print(self.json)
            raise

    @property
    def all_concepts(self):
        return self.dataframe.loc[:, 'concept.conceptPath'].unique()

    def aggregate_numeric_on_trial_visit(self, concept):
        concept_groups = self.dataframe.groupby('concept.conceptPath')
        concept_group = concept_groups.get_group(concept)
        columns_of_interest = ['patient.inTrialId', 'trial visit.relTimeLabel', 'numericValue']
        df_subset = concept_group.loc[:, columns_of_interest]
        df_subset.columns = ['subject', 'visit_label', 'value']
        return df_subset.pivot_table(index='visit_label',
                                     values='value',
                                     columns='subject',
                                     aggfunc='mean')


class ObservationSetHD:

    def __init__(self, json):
        self.json = json
        try:
            self.dataframe = json_normalize(_format_observations(self.json))
            assert not self.dataframe.shape == (0, 0)
        except:
            print(self.json)
            raise

    @property
    def all_biomarkers(self):
        return self.dataframe.groupby(['biomarker.biomarker', 'biomarker.label']).size()

    def biomarker_boxplot(self):
        if len(self.all_biomarkers) > 1:
            return self.dataframe.boxplot(column='numericValue', by=['biomarker.biomarker', 'biomarker.label'])


class Studies:

    def __init__(self, json):
        self.json = json
        self.dataframe = json_normalize(json['studies'])


class StudyList:

    def __init__(self, study_list):
        for study_id in study_list:
            self.__dict__[study_id.replace('-', '_')] = study_id


class TreeNodes:

    def __init__(self, json):
        self.json = json
        self.dataframe = json_normalize(self.create_list())

    def __repr__(self):
        return self.pretty()

    def pretty(self, root=None, depth=0, spacing=2):
        """
        Create a pretty representation of tree.
        """
        
        if root is None:
            top_nodes = [self.pretty(root, depth=0, spacing=spacing) for root in self.json.get('tree_nodes')]
            return '\n'.join(top_nodes)

        fmt = "%s%s/" if root.get('children') else "%s%s"
        s = fmt % (" " * depth * spacing, "{}  ({})".format(root.get('name'), root.get('patientCount')))
        for child in root.get('children', {}):
            s += "\n%s" % self.pretty(child, depth + 1, spacing)
        return s
    
    def create_list(self, root=None):
        """
        Create dataframe representation of tree.
        """
        
        if root is None:
            top_nodes = []
            for root in self.json.get('tree_nodes'):
                top_nodes += self.create_list(root)
            return top_nodes

        thisnode = root.copy()
        thisnode.pop("children", None)
        s = [thisnode]

        for child in root.get('children', {}):
            s += self.create_list(child)
        
        return s
        


class PatientSets:

    def __init__(self, json):
        self.json = json
        self.dataframe = json_normalize(json.get('patientSets'))

